@@ -5,7 +5,7 @@ use strict;
 use utf8;
 use Carp;
 
-our $VERSION = 'v0.1.1';
+our $VERSION = 'v0.1.2';
 
 use Export::Attrs;
 use Scalar::Util qw( weaken );
@@ -43,6 +43,7 @@ eval {
 };
 
 
+# FIXME make ff() reentrant
 sub ff :Export(:DEFAULT) {
     my ($dur) = @_;
 
@@ -51,23 +52,27 @@ sub ff :Export(:DEFAULT) {
         $a->{id} cmp $b->{id}   # preserve order to simplify tests
         } @Timers;
     my $next_at = @Timers ? $Timers[0]{start}+$Timers[0]{after} : 0;
+    $next_at = sprintf '%.6f', $next_at;
 
     if (!defined $dur) {
         $dur = $next_at > $Relative ? $next_at - $Relative : 0;
+        $dur = sprintf '%.6f', $dur;
     }
     croak "ff($dur): negative time not invented yet" if $dur < 0;
 
-    if (!$next_at || $next_at > $Relative+$dur) {
+    if ($next_at == 0 || $next_at > $Relative+$dur) {
         $Relative += $dur;
+        $Relative = sprintf '%.6f', $Relative;
         return;
     }
 
     if ($next_at > $Relative) {
         $dur -= $next_at - $Relative;
+        $dur = sprintf '%.6f', $dur;
         $Relative = $next_at;
     }
-    $Timers[0]{cb}->();
-    if (!$Timers[0]{repeat}) {
+    my $cb = $Timers[0]{cb};
+    if ($Timers[0]{repeat} == 0) {
         if ($Timers[0]{watcher}) {
             _stop_timer($Timers[0]{watcher});
         }
@@ -79,6 +84,7 @@ sub ff :Export(:DEFAULT) {
         $Timers[0]{after} = $Timers[0]{repeat};
         $Timers[0]{start} = $Relative;
     }
+    $cb->();
     @_ = ($dur);
     goto &ff;
 }
@@ -91,8 +97,8 @@ sub _add_timer {
         id      => $id,
         start   => $Relative,
         loop    => $loop,
-        after   => $after,
-        repeat  => $repeat,
+        after   => sprintf('%.6f', $after < 0 ? 0 : $after),
+        repeat  => sprintf('%.6f', $repeat < 0 ? 0 : $repeat),
         cb      => $cb,
         watcher => $watcher,
     };
@@ -139,6 +145,7 @@ sub _mock_core_global {
         my $dur = int $_[0];
         croak 'sleep with negative value is not supported' if $dur < 0;
         $Relative += $dur;
+        $Relative = sprintf '%.6f', $Relative;
         return $dur;
     });
     return;
@@ -152,7 +159,7 @@ sub _mock_time_hires {
 
     $Module{'Time::HiRes'} = Test::MockModule->new('Time::HiRes');
     $Module{'Time::HiRes'}->mock(time => sub () {
-        return $Absolute + $Relative;
+        return 0+sprintf '%.6f', $Absolute + $Relative;
     });
     $Module{'Time::HiRes'}->mock(gettimeofday => sub () {
         my $t = sprintf '%.6f', $Absolute + $Relative;
@@ -161,10 +168,10 @@ sub _mock_time_hires {
     $Module{'Time::HiRes'}->mock(clock_gettime => sub (;$) {
         my ($which) = @_;
         if ($which == CLOCK_REALTIME()) {
-            return $Absolute + $Relative;
+            return 0+sprintf '%.6f', $Absolute + $Relative;
         }
         elsif ($which == CLOCK_MONOTONIC()) {
-            return $Monotonic + $Relative;
+            return 0+sprintf '%.6f', $Monotonic + $Relative;
         }
         return TIME_HIRES_CLOCK_NOT_SUPPORTED;
     });
@@ -180,18 +187,21 @@ sub _mock_time_hires {
         croak 'sleep without arg is not supported' if !@_;
         croak "Time::HiRes::sleep($seconds): negative time not invented yet" if $seconds < 0;
         $Relative += $seconds;
+        $Relative = sprintf '%.6f', $Relative;
         return $seconds;
     });
     $Module{'Time::HiRes'}->mock(usleep => sub ($) {
         my ($useconds) = @_;
         croak "Time::HiRes::usleep($useconds): negative time not invented yet" if $useconds < 0;
         $Relative += $useconds / MICROSECONDS;
+        $Relative = sprintf '%.6f', $Relative;
         return $useconds;
     });
     $Module{'Time::HiRes'}->mock(nanosleep => sub ($) {
         my ($nanoseconds) = @_;
         croak "Time::HiRes::nanosleep($nanoseconds): negative time not invented yet" if $nanoseconds < 0;
         $Relative += $nanoseconds / NANOSECONDS;
+        $Relative = sprintf '%.6f', $Relative;
         return $nanoseconds;
     });
     $Module{'Time::HiRes'}->mock(clock_nanosleep => sub ($$;$) {
@@ -200,6 +210,7 @@ sub _mock_time_hires {
         croak 'only CLOCK_REALTIME and CLOCK_MONOTONIC are supported' if $which != CLOCK_REALTIME() && $which != CLOCK_MONOTONIC();
         croak 'only flags=0 is supported' if $flags;
         $Relative += $nanoseconds / NANOSECONDS;
+        $Relative = sprintf '%.6f', $Relative;
         return $nanoseconds;
     });
     return;
@@ -214,10 +225,10 @@ sub _mock_ev { ## no critic (ProhibitExcessComplexity)
     $Module{'EV::Timer'}    = Test::MockModule->new('EV::Timer',    no_auto=>1);
     $Module{'EV::Periodic'} = Test::MockModule->new('EV::Periodic', no_auto=>1);
     $Module{'EV'}->mock(time => sub () {
-        return $Absolute + $Relative;
+        return 0+sprintf '%.6f', $Absolute + $Relative;
     });
     $Module{'EV'}->mock(now => sub () {
-        return $Absolute + $Relative;
+        return 0+sprintf '%.6f', $Absolute + $Relative;
     });
     $Module{'EV'}->mock(sleep => sub ($) {
         my ($seconds) = @_;
@@ -225,6 +236,7 @@ sub _mock_ev { ## no critic (ProhibitExcessComplexity)
             $seconds = 0;
         }
         $Relative += $seconds;
+        $Relative = sprintf '%.6f', $Relative;
         return;
     });
     $Module{'EV'}->mock(run => sub (;$) {
@@ -232,9 +244,23 @@ sub _mock_ev { ## no critic (ProhibitExcessComplexity)
         my $tick = 0;
         my $w;
         if (@Timers) {
+            ## no critic (ProhibitBooleanGrep)
             $w = $Module{'EV'}->original('timer')->(
-                $WAIT_ONE_TICK, $WAIT_ONE_TICK, sub { if (!$tick++) { ff() } }
+                $WAIT_ONE_TICK, $WAIT_ONE_TICK, sub {
+                    if (!$tick++ || !$flags) {
+                        ff();
+                    }
+                    if (!@Timers) {
+                        shift->stop;
+                    }
+                    elsif (!$flags && !grep {$_} map {$_->keepalive} grep {$_} map {$_->{watcher}} @Timers) {
+                        shift->keepalive(0);
+                    }
+                }
             );
+            if (!$flags && !grep {$_} map {$_->keepalive} grep {$_} map {$_->{watcher}} @Timers) {
+                $w->keepalive(0);
+            }
         }
         # $tick above and this second RUN_ONCE is work around bug in EV-4.10+
         # http://lists.schmorp.de/pipermail/libev/2016q1/002656.html
@@ -247,14 +273,12 @@ sub _mock_ev { ## no critic (ProhibitExcessComplexity)
         if ($flags && $flags == EV::RUN_ONCE()) {
             $Module{'EV'}->original('run')->(@_);
         }
-        # FIXME returned $active value doesn't take in account @Timers
         return $Module{'EV'}->original('run')->(@_);
     });
     $Module{'EV'}->mock(timer => sub ($$$) {
         my ($after, $repeat, $cb) = @_;
         my $w = $Module{'EV'}->original('timer_ns')->(@_);
-        my $weakw = $w;
-        weaken($weakw);
+        weaken(my $weakw = $w);
         _add_timer('EV', $after, $repeat, sub { $weakw && $weakw->invoke(EV::TIMER()) }, $w);
         return $w;
     });
@@ -266,13 +290,17 @@ sub _mock_ev { ## no critic (ProhibitExcessComplexity)
     $Module{'EV'}->mock(periodic => sub ($$$$) {
         my ($at, $repeat, $reschedule_cb, $cb) = @_;
         croak 'reschedule_cb is not supported yet' if $reschedule_cb;
-        if ($repeat && $at < $Absolute+$Relative) {
-            $at += $repeat * int(($Absolute+$Relative - $at) / $repeat + 1);
+        $at = sprintf '%.6f', $at < 0 ? 0 : $at;
+        $repeat = sprintf '%.6f', $repeat < 0 ? 0 : $repeat;
+        my $now = sprintf '%.6f', $Absolute + $Relative;
+        if ($repeat > 0 && $at < $now) {
+            $at += $repeat * int(($now - $at) / $repeat + 1);
+            $at = sprintf '%.6f', $at;
         }
-        my $after = $at > $Absolute+$Relative ? $at - ($Absolute+$Relative) : 0;
+        my $after = $at > $now ? $at - $now : 0;
+        $after = sprintf '%.6f', $after;
         my $w = $Module{'EV'}->original('periodic_ns')->(@_);
-        my $weakw = $w;
-        weaken($weakw);
+        weaken(my $weakw = $w);
         _add_timer('EV', $after, $repeat, sub { $weakw && $weakw->invoke(EV::TIMER()) }, $w);
         return $w;
     });
@@ -311,8 +339,8 @@ sub _mock_ev { ## no critic (ProhibitExcessComplexity)
         my ($timer) = grep { $_->{watcher} && $_->{watcher} eq $w } @Timers, @Timers_ns;
         if ($timer) {
             $timer->{start} = $Relative;
-            $timer->{after} = $after;
-            $timer->{repeat}= $repeat;
+            $timer->{after} = sprintf '%.6f', $after < 0 ? 0 : $after;
+            $timer->{repeat}= sprintf '%.6f', $repeat < 0 ? 0 : $repeat;
         }
         return;
     });
@@ -320,17 +348,20 @@ sub _mock_ev { ## no critic (ProhibitExcessComplexity)
         my ($w) = @_;
         my ($timer) = grep { $_->{watcher} && $_->{watcher} eq $w } @Timers, @Timers_ns;
         if ($timer) {
-            return $timer->{start} + $timer->{after} - $Relative;
+            return 0+sprintf '%.6f', $timer->{start} + $timer->{after} - $Relative;
         }
         return;
     });
     $Module{'EV::Timer'}->mock(again => sub {
         my ($w, $repeat) = @_;
+        if (defined $repeat && $repeat < 0) {
+            $repeat = 0;
+        }
         my ($active) = grep { $_->{watcher} && $_->{watcher} eq $w } @Timers;
         my ($inactive) = grep { $_->{watcher} && $_->{watcher} eq $w } @Timers_ns;
         if ($active) {
-            $active->{repeat} = $repeat // $active->{repeat};
-            if ($active->{repeat}) {
+            $active->{repeat} = sprintf '%.6f', $repeat // $active->{repeat};
+            if ($active->{repeat} > 0) {
                 $active->{after} = $active->{repeat};
                 $active->{start} = $Relative;
             }
@@ -339,8 +370,8 @@ sub _mock_ev { ## no critic (ProhibitExcessComplexity)
             }
         }
         elsif ($inactive) {
-            $inactive->{repeat} = $repeat // $inactive->{repeat};
-            if ($inactive->{repeat}) {
+            $inactive->{repeat} = sprintf '%.6f', $repeat // $inactive->{repeat};
+            if ($inactive->{repeat} > 0) {
                 $inactive->{after} = $inactive->{repeat};
                 $inactive->{start} = $Relative;
                 _start_timer($inactive->{watcher});
@@ -363,10 +394,15 @@ sub _mock_ev { ## no critic (ProhibitExcessComplexity)
     $Module{'EV::Periodic'}->mock(set => sub {
         my ($w, $at, $repeat, $reschedule_cb, $cb) = @_;
         croak 'reschedule_cb is not supported yet' if $reschedule_cb;
-        if ($repeat && $at < $Absolute+$Relative) {
-            $at += $repeat * int(($Absolute+$Relative - $at) / $repeat + 1);
+        $at = sprintf '%.6f', $at < 0 ? 0 : $at;
+        $repeat = sprintf '%.6f', $repeat < 0 ? 0 : $repeat;
+        my $now = sprintf '%.6f', $Absolute + $Relative;
+        if ($repeat > 0 && $at < $now) {
+            $at += $repeat * int(($now - $at) / $repeat + 1);
+            $at = sprintf '%.6f', $at;
         }
-        my $after = $at > $Absolute+$Relative ? $at - ($Absolute+$Relative) : 0;
+        my $after = $at > $now ? $at - $now : 0;
+        $after = sprintf '%.6f', $after;
         my ($timer) = grep { $_->{watcher} && $_->{watcher} eq $w } @Timers, @Timers_ns;
         if ($timer) {
             $timer->{start} = $Relative;
@@ -382,7 +418,7 @@ sub _mock_ev { ## no critic (ProhibitExcessComplexity)
         my ($w) = @_;
         my ($timer) = grep { $_->{watcher} && $_->{watcher} eq $w } @Timers, @Timers_ns;
         if ($timer) {
-            return $timer->{start} + $timer->{after};
+            return 0+sprintf '%.6f', $timer->{start} + $timer->{after};
         }
         return;
     });
@@ -458,8 +494,7 @@ Test::Mock::Time - Deterministic time & timers for event loop tests
 
 =head1 VERSION
 
-This document describes Test::Mock::Time version v0.1.1
-=head1 SYNOPSIS
+This document describes Test::Mock::Time version v0.1.2=head1 SYNOPSIS
 
   use Test::Mock::Time;
 
@@ -595,7 +630,6 @@ All required methods except:
 
     EV::once
     EV::Watcher::feed_event
-    EV::Watcher::keepalive
 
 =back
 
